module Util

export savedata

# Path Management

"""
    Context{P}

Execution context for an experiment.
"""
struct Context{P}
    indir::String
    outdir::String
    params::P
    # TODO: get rid of this when latex output does no longer depend on path
    path::String
end

Context(root::String; params...) = Context(root, root, NamedTuple(params), "")
Context(inroot::String, outroot::String; params...) =
    Context(inroot, outroot, NamedTuple(params), "")
Context(ctx::Context, path::String; params...) =
    Context(joinpath(ctx.indir, path), joinpath(ctx.outdir, path),
        merge(ctx.params, NamedTuple(params)),
        joinpath(ctx.path, path))
Context(ctx::Context; params...) =
    Context(ctx.indir, ctx.outdir,
        merge(ctx.params, NamedTuple(params)),
        ctx.path)

"""
    (ctx::Context)(f, subdir; params...)

Execute experiment `f` on a context relative to `ctx` and `subdir` using
parameters `params`.
"""
function (ctx::Context)(f::Function, subdir::String; params...)
    # maybe debug output with params?
    subctx = Context(ctx, subdir; params...)
    mkpath(subctx.outdir)
    println("starting experiment `$(subctx.path)` ...")
    f(subctx)
end

# LaTeX Data Output

savedata(filename; data...) =
    open(filename; write=true) do io
        _savedata(io; data...)
    end

function _savedata(io; data...)
    isvalidname(x) = contains(string(x), r"^[^/\\]+$")

    # define tex command to access data
    texcmd = "Data"
    write(io, """
        \\providecommand{\\$(texcmd)}[1]{
            \\csname $(texcmd)/#1\\endcsname
        }
        """)

    for (key, value) in pairs(data)
        isvalidname(key) && isvalidname(value) ||
            throw(ArgumentError("invalid key/value: $(key => value)"))

        # actual data encoded as tex-commands
        write(io, "\\expandafter\\def")
        write(io, "\\csname $(texcmd)/\\DataPrefix/$(key)\\endcsname{")
        show(io, MIME("text/latex"), value)
        write(io, "}\n")
    end
end

# LaTex special printing rules

Base.show(io, ::MIME"text/latex", x) = print(io, x)

function Base.show(io::IO, ::MIME"text/latex", x::Union{Integer,AbstractFloat})
    print(io, "\\pgfmathprintnumber{", x ,"}")
end

end
